#pragma once

#include <chrono>
#include <rewindow/types/key.h>
#include <rewindow/types/mousebutton.h>

namespace ReWindow {

    using namespace std::chrono_literals;
    using precision_clock = std::chrono::high_resolution_clock;
    using precision_timestamp = precision_clock::time_point;

    class TimestampedEvent {
    public:
        precision_timestamp Timestamp;
        explicit TimestampedEvent(precision_timestamp timestamp) : Timestamp(timestamp) {}
        TimestampedEvent() : Timestamp(precision_clock::now())
        { }
    };

    class RWindowEvent : public TimestampedEvent {
    public:
        RWindowEvent() : TimestampedEvent() { }
    };

    class InputDeviceEvent : public RWindowEvent {
        public:
        InputDeviceEvent() : RWindowEvent() { }
    };

    class KeyboardEvent : public InputDeviceEvent {
    public:
        Key key;
        KeyState state;
        KeyboardEvent(const Key& key, KeyState state) : InputDeviceEvent(), key(key), state(state) {}
    };
    class MouseEvent : public InputDeviceEvent {};

    class GamepadEvent : public InputDeviceEvent {};

    class MouseButtonEvent : public MouseEvent {
    public:
        MouseButton Button;
        bool Pressed; // TODO: replace with MouseButtonState enum
        MouseButtonEvent() = default;
        MouseButtonEvent(const MouseButton& button, bool pressed) : MouseEvent(), Button(button), Pressed(pressed) {}
    };

    class MouseButtonDownEvent : public MouseButtonEvent {
    public:
        MouseButton Button;

        MouseButtonDownEvent() = default;
        MouseButtonDownEvent(const MouseButton& button) : MouseButtonEvent(button, true) {}
    };

    class MouseButtonUpEvent : public MouseButtonEvent {
    public:

        MouseButtonUpEvent() = default;
        MouseButtonUpEvent(const MouseButton& button) : MouseButtonEvent(button, true) {}
    };

    class MouseMoveEvent : public MouseEvent {
    public:
        Vector2 Position;
        Vector2 Delta;
        MouseMoveEvent(const Vector2 &pos) : MouseEvent(), Position(pos)
        {}
        MouseMoveEvent(int x, int y) : MouseEvent(), Position(Vector2(x, y))
        {}
    };

    class KeyDownEvent : public KeyboardEvent {
    public:
        explicit KeyDownEvent(const Key& key) : KeyboardEvent(key, KeyState::Pressed) {}
    };

    class KeyUpEvent : public KeyboardEvent {
    public:
        explicit KeyUpEvent(const Key& key) : KeyboardEvent(key, KeyState::Released) {}
    };

    class MouseWheelEvent : public MouseEvent
    {
    public:
        int WheelMovement;
        MouseWheelEvent() = default;
        MouseWheelEvent(int wheel) : MouseEvent(), WheelMovement(wheel) {}
    };



    class WindowResizeRequestEvent : public MouseButtonEvent
    {
    public:
        Vector2 Size;
    };
}